extern int v3p_netlib_zlaset_(
  char *uplo,
  v3p_netlib_integer *m,
  v3p_netlib_integer *n,
  v3p_netlib_doublecomplex *alpha,
  v3p_netlib_doublecomplex *beta,
  v3p_netlib_doublecomplex *a,
  v3p_netlib_integer *lda,
  v3p_netlib_ftnlen uplo_len
  );
