/* This header file mangles names added to the namespace by v3p_f2c.h.
   There are purposely no include blockers so translation units can
   mangle/unmangle repeatedly.  */

/* Functions */
#define F77_aloc v3p_netlib_F77_aloc
#define exit_ v3p_netlib_exit_
#define c_abs v3p_netlib_c_abs
#define c_div v3p_netlib_c_div
#define c_sqrt v3p_netlib_c_sqrt
#define d_cnjg v3p_netlib_d_cnjg
#define d_imag v3p_netlib_d_imag
#define d_lg10 v3p_netlib_d_lg10
#define d_sign v3p_netlib_d_sign
#define f__cabs v3p_netlib_f__cabs
#define i_dnnt v3p_netlib_i_dnnt
#define pow_dd v3p_netlib_pow_dd
#define pow_di v3p_netlib_pow_di
#define pow_ii v3p_netlib_pow_ii
#define pow_ri v3p_netlib_pow_ri
#define r_cnjg v3p_netlib_r_cnjg
#define r_imag v3p_netlib_r_imag
#define r_sign v3p_netlib_r_sign
#define s_cat v3p_netlib_s_cat
#define s_cmp v3p_netlib_s_cmp
#define s_copy v3p_netlib_s_copy
#define sig_die v3p_netlib_sig_die
#define z_abs v3p_netlib_z_abs
#define z_div v3p_netlib_z_div
#define z_sqrt v3p_netlib_z_sqrt

/* Types */
#define C_f v3p_netlib_C_f
#define C_fp v3p_netlib_C_fp
#define E_f v3p_netlib_E_f
#define E_fp v3p_netlib_E_fp
#define H_f v3p_netlib_H_f
#define H_fp v3p_netlib_H_fp
#define I_fp v3p_netlib_I_fp
#define J_fp v3p_netlib_J_fp
#define K_fp v3p_netlib_K_fp
#define L_fp v3p_netlib_L_fp
#define Multitype v3p_netlib_Multitype
#define Namelist v3p_netlib_Namelist
#define R_fp v3p_netlib_R_fp
#define S_fp v3p_netlib_S_fp
#define U_fp v3p_netlib_U_fp
#define Vardesc v3p_netlib_Vardesc
#define Z_f v3p_netlib_Z_f
#define Z_fp v3p_netlib_Z_fp
#define address v3p_netlib_address
#define alist v3p_netlib_alist
#define cilist v3p_netlib_cilist
#define cllist v3p_netlib_cllist
#define complex v3p_netlib_complex
#define doublecomplex v3p_netlib_doublecomplex
#define doublereal v3p_netlib_doublereal
#define flag v3p_netlib_flag
#define ftnint v3p_netlib_ftnint
#define ftnlen v3p_netlib_ftnlen
#define icilist v3p_netlib_icilist
#define inlist v3p_netlib_inlist
#define integer v3p_netlib_integer
#define integer1 v3p_netlib_integer1
#define logical v3p_netlib_logical
#define logical1 v3p_netlib_logical1
#define longint v3p_netlib_longint
#define olist v3p_netlib_olist
#define real v3p_netlib_real
#define shortint v3p_netlib_shortint
#define shortlogical v3p_netlib_shortlogical
#define uinteger v3p_netlib_uinteger
#define ulongint v3p_netlib_ulongint
