extern int v3p_netlib_zgges_(
  char v3p_netlib_const *jobvsl,
  char v3p_netlib_const *jobvsr,
  char v3p_netlib_const *sort,
  v3p_netlib_logical (*selctg)(v3p_netlib_doublecomplex*,
                               v3p_netlib_doublecomplex*),
  v3p_netlib_integer *n,
  v3p_netlib_doublecomplex *a,
  v3p_netlib_integer *lda,
  v3p_netlib_doublecomplex *b,
  v3p_netlib_integer *ldb,
  v3p_netlib_integer *sdim,
  v3p_netlib_doublecomplex *alpha,
  v3p_netlib_doublecomplex *beta,
  v3p_netlib_doublecomplex *vsl,
  v3p_netlib_integer *ldvsl,
  v3p_netlib_doublecomplex *vsr,
  v3p_netlib_integer *ldvsr,
  v3p_netlib_doublecomplex *work,
  v3p_netlib_integer *lwork,
  v3p_netlib_doublereal *rwork,
  v3p_netlib_logical *bwork,
  v3p_netlib_integer *info,
  v3p_netlib_ftnlen jobvsl_len,
  v3p_netlib_ftnlen jobvsr_len,
  v3p_netlib_ftnlen sort_len
  );
