extern int v3p_netlib_dnppla_(
  void (*op)(v3p_netlib_integer v3p_netlib_const*,
             v3p_netlib_integer v3p_netlib_const*,
             v3p_netlib_doublereal*,
             v3p_netlib_doublereal v3p_netlib_const*),
  void (*iovect)(v3p_netlib_integer v3p_netlib_const*,
                 v3p_netlib_integer v3p_netlib_const*,
                 v3p_netlib_doublereal*,
                 v3p_netlib_integer v3p_netlib_const*,
                 v3p_netlib_integer v3p_netlib_const*),
  v3p_netlib_integer *n,
  v3p_netlib_integer *nperm,
  v3p_netlib_integer *nop,
  v3p_netlib_integer *nmval,
  v3p_netlib_doublereal *val,
  v3p_netlib_integer *nmvec,
  v3p_netlib_doublereal *vec,
  v3p_netlib_integer *nblock,
  v3p_netlib_doublereal *h__,
  v3p_netlib_doublereal *hv,
  v3p_netlib_doublereal *p,
  v3p_netlib_doublereal *q,
  v3p_netlib_doublereal *bound,
  v3p_netlib_doublereal *d__,
  v3p_netlib_doublereal *delta,
  v3p_netlib_logical *small,
  v3p_netlib_logical *raritz,
  v3p_netlib_doublereal *eps
  );
