extern int v3p_netlib_dsaup2_(
  v3p_netlib_integer *ido,
  char *bmat,
  v3p_netlib_integer *n,
  char *which,
  v3p_netlib_integer *nev,
  v3p_netlib_integer *np,
  v3p_netlib_doublereal *tol,
  v3p_netlib_doublereal *resid,
  v3p_netlib_integer *mode,
  v3p_netlib_integer *iupd,
  v3p_netlib_integer *ishift,
  v3p_netlib_integer *mxiter,
  v3p_netlib_doublereal *v,
  v3p_netlib_integer *ldv,
  v3p_netlib_doublereal *h__,
  v3p_netlib_integer *ldh,
  v3p_netlib_doublereal *ritz,
  v3p_netlib_doublereal *bounds,
  v3p_netlib_doublereal *q,
  v3p_netlib_integer *ldq,
  v3p_netlib_doublereal *workl,
  v3p_netlib_integer *ipntr,
  v3p_netlib_doublereal *workd,
  v3p_netlib_integer *info,
  v3p_netlib_ftnlen bmat_len,
  v3p_netlib_ftnlen which_len
  );
