extern int v3p_netlib_zgecon_(
  char *norm,
  v3p_netlib_integer *n,
  v3p_netlib_doublecomplex *a,
  v3p_netlib_integer *lda,
  v3p_netlib_doublereal *anorm,
  v3p_netlib_doublereal *rcond,
  v3p_netlib_doublecomplex *work,
  v3p_netlib_doublereal *rwork,
  v3p_netlib_integer *info,
  v3p_netlib_ftnlen norm_len
  );
