/*: Finds a few eigenvalues and eigenvectors at either end of the spectrum of a large sparse symmetric matrix.  */
extern int v3p_netlib_dnlaso_(
  void (*op)(v3p_netlib_integer v3p_netlib_const*,
             v3p_netlib_integer v3p_netlib_const*,
             v3p_netlib_doublereal v3p_netlib_const*,
             v3p_netlib_doublereal*),
  void (*iovect)(v3p_netlib_integer v3p_netlib_const*,
                 v3p_netlib_integer v3p_netlib_const*,
                 v3p_netlib_doublereal*,
                 v3p_netlib_integer v3p_netlib_const*,
                 v3p_netlib_integer v3p_netlib_const*),
  v3p_netlib_integer v3p_netlib_const *n,
  v3p_netlib_integer v3p_netlib_const *nval,
  v3p_netlib_integer v3p_netlib_const *nfig,
  v3p_netlib_integer v3p_netlib_const *nperm,
  v3p_netlib_integer v3p_netlib_const *nmval,
  v3p_netlib_doublereal *val,
  v3p_netlib_integer v3p_netlib_const *nmvec,
  v3p_netlib_doublereal *vec,
  v3p_netlib_integer v3p_netlib_const *nblock,
  v3p_netlib_integer v3p_netlib_const *maxop,
  v3p_netlib_integer v3p_netlib_const *maxj,
  v3p_netlib_doublereal *work,
  v3p_netlib_integer *ind,
  v3p_netlib_integer *ierr
  );
extern int v3p_netlib_dnwla_(
  void (*op)(v3p_netlib_integer*,
             v3p_netlib_integer*,
             v3p_netlib_doublereal*,
             v3p_netlib_doublereal*),
  void (*iovect)(v3p_netlib_integer*,
                 v3p_netlib_integer*,
                 v3p_netlib_doublereal*,
                 v3p_netlib_integer*,
                 v3p_netlib_integer*),
  v3p_netlib_integer *n,
  v3p_netlib_integer *nband,
  v3p_netlib_integer *nval,
  v3p_netlib_integer *nfig,
  v3p_netlib_integer *nperm,
  v3p_netlib_doublereal *val,
  v3p_netlib_integer *nmvec,
  v3p_netlib_doublereal *vec,
  v3p_netlib_integer *nblock,
  v3p_netlib_integer *maxop,
  v3p_netlib_integer *maxj,
  v3p_netlib_integer *nop,
  v3p_netlib_doublereal *p1,
  v3p_netlib_doublereal *p0,
  v3p_netlib_doublereal *res,
  v3p_netlib_doublereal *tau,
  v3p_netlib_doublereal *otau,
  v3p_netlib_doublereal *t,
  v3p_netlib_doublereal *alp,
  v3p_netlib_doublereal *bet,
  v3p_netlib_doublereal *s,
  v3p_netlib_doublereal *p2,
  v3p_netlib_doublereal *bound,
  v3p_netlib_doublereal *atemp,
  v3p_netlib_doublereal *vtemp,
  v3p_netlib_doublereal *d__,
  v3p_netlib_integer *ind,
  v3p_netlib_logical *small,
  v3p_netlib_logical *raritz,
  v3p_netlib_doublereal *delta,
  v3p_netlib_doublereal *eps,
  v3p_netlib_integer *ierr
  );
