extern int v3p_netlib_ztgsy2_(
  char *trans,
  v3p_netlib_integer *ijob,
  v3p_netlib_integer *m,
  v3p_netlib_integer *n,
  v3p_netlib_doublecomplex *a,
  v3p_netlib_integer *lda,
  v3p_netlib_doublecomplex *b,
  v3p_netlib_integer *ldb,
  v3p_netlib_doublecomplex *c__,
  v3p_netlib_integer *ldc,
  v3p_netlib_doublecomplex *d__,
  v3p_netlib_integer *ldd,
  v3p_netlib_doublecomplex *e,
  v3p_netlib_integer *lde,
  v3p_netlib_doublecomplex *f,
  v3p_netlib_integer *ldf,
  v3p_netlib_doublereal *scale,
  v3p_netlib_doublereal *rdsum,
  v3p_netlib_doublereal *rdscal,
  v3p_netlib_integer *info,
  v3p_netlib_ftnlen trans_len
  );
