extern int v3p_netlib_drot_(
  v3p_netlib_integer *n,
  v3p_netlib_doublereal *dx,
  v3p_netlib_integer *incx,
  v3p_netlib_doublereal *dy,
  v3p_netlib_integer *incy,
  v3p_netlib_doublereal *c__,
  v3p_netlib_doublereal *s
  );
