extern int v3p_netlib_zlatrs_(
  char *uplo,
  char *trans,
  char *diag,
  char *normin,
  v3p_netlib_integer *n,
  v3p_netlib_doublecomplex *a,
  v3p_netlib_integer *lda,
  v3p_netlib_doublecomplex *x,
  v3p_netlib_doublereal *scale,
  v3p_netlib_doublereal *cnorm,
  v3p_netlib_integer *info,
  v3p_netlib_ftnlen uplo_len,
  v3p_netlib_ftnlen trans_len,
  v3p_netlib_ftnlen diag_len,
  v3p_netlib_ftnlen normin_len
  );
