/* This header undefines mangled names defined by v3p_f2c_mangle.h
   There are purposely no include blockers so translation units can
   mangle/unmangle repeatedly.  */

/* Functions */
#undef F77_aloc
#undef exit_
#undef c_abs
#undef c_div
#undef c_sqrt
#undef d_cnjg
#undef d_imag
#undef d_lg10
#undef d_sign
#undef f__cabs
#undef i_dnnt
#undef pow_dd
#undef pow_di
#undef pow_ii
#undef pow_ri
#undef r_cnjg
#undef r_imag
#undef r_sign
#undef s_cat
#undef s_cmp
#undef s_copy
#undef sig_die
#undef z_abs
#undef z_div
#undef z_sqrt

/* Types */
#undef C_f
#undef C_fp
#undef E_f
#undef E_fp
#undef H_f
#undef H_fp
#undef I_fp
#undef J_fp
#undef K_fp
#undef L_fp
#undef Multitype
#undef Namelist
#undef R_fp
#undef S_fp
#undef U_fp
#undef Vardesc
#undef Z_f
#undef Z_fp
#undef address
#undef alist
#undef cilist
#undef cllist
#undef complex
#undef doublecomplex
#undef doublereal
#undef flag
#undef ftnint
#undef ftnlen
#undef icilist
#undef inlist
#undef integer
#undef integer1
#undef logical
#undef logical1
#undef longint
#undef olist
#undef real
#undef shortint
#undef shortlogical
#undef uinteger
#undef ulongint
