extern int v3p_netlib_dspr_(
  char *uplo,
  v3p_netlib_integer *n,
  v3p_netlib_doublereal *alpha,
        v3p_netlib_doublereal *x,
  v3p_netlib_integer *incx,
  v3p_netlib_doublereal *ap,
  v3p_netlib_ftnlen uplo_len
  );

