extern int v3p_netlib_ztgsen_(
  v3p_netlib_integer *ijob,
  v3p_netlib_logical *wantq,
  v3p_netlib_logical *wantz,
  v3p_netlib_logical *select,
  v3p_netlib_integer *n,
  v3p_netlib_doublecomplex *a,
  v3p_netlib_integer *lda,
  v3p_netlib_doublecomplex *b,
  v3p_netlib_integer *ldb,
  v3p_netlib_doublecomplex *alpha,
  v3p_netlib_doublecomplex *beta,
  v3p_netlib_doublecomplex *q,
  v3p_netlib_integer *ldq,
  v3p_netlib_doublecomplex *z__,
  v3p_netlib_integer *ldz,
  v3p_netlib_integer *m,
  v3p_netlib_doublereal *pl,
  v3p_netlib_doublereal *pr,
  v3p_netlib_doublereal *dif,
  v3p_netlib_doublecomplex *work,
  v3p_netlib_integer *lwork,
  v3p_netlib_integer *iwork,
  v3p_netlib_integer *liwork,
  v3p_netlib_integer *info
  );
