/* This header undefines mangled names defined by v3p_
   There are purposely no include blockers so translation units can
   mangle/unmangle repeatedly. */

/* Functions */

#undef active_
#undef adaptquad_
#undef balanc_
#undef balbak_
#undef bmv_
#undef calcsc_
#undef cauchy_
#undef caxpy_
#undef ccopy_
#undef cdiv_
#undef cdotc_
#undef cg_
#undef chscdf_
#undef cmprlb_
#undef cqrdc_
#undef cqrsl_
#undef cscal_
#undef csrot_
#undef csvdc_
#undef cswap_
#undef cub_
#undef dasum_
#undef daxpy_
#undef dcabs1_
#undef dchscdf_
#undef dcopy_
#undef dcsrch_
#undef dcstep_
#undef ddot_
#undef dgecon_
#undef dgemm_
#undef dgemv_
#undef dgeqr2_
#undef dgeqrf_
#undef dger_
#undef dgerq2_
#undef dgesc2_
#undef dgetc2_
#undef dgetv0_
#undef dggbak_
#undef dggbal_
#undef dgges_
#undef dgghrd_
#undef dgpfa2f_
#undef dgpfa3f_
#undef dgpfa5f_
#undef dgpfa_
#undef dhgeqz_
#undef dlabad_
#undef dlabax_
#undef dlabcm_
#undef dlabfc_
#undef dlacon_
#undef dlacpy_
#undef dladiv_
#undef dlae2_
#undef dlaeig_
#undef dlaev2_
#undef dlag2_
#undef dlager_
#undef dlagv2_
#undef dlamc1_
#undef dlamc2_
#undef dlamc3_
#undef dlamc4_
#undef dlamc5_
#undef dlamch_
#undef dlange_
#undef dlanhs_
#undef dlanst_
#undef dlapy2_
#undef dlapy3_
#undef dlaran_
#undef dlarf_
#undef dlarfb_
#undef dlarfg_
#undef dlarft_
#undef dlarnv_
#undef dlartg_
#undef dlaruv_
#undef dlascl_
#undef dlaset_
#undef dlasr_
#undef dlasrt_
#undef dlassq_
#undef dlasv2_
#undef dlaswp_
#undef dlatdf_
#undef dlatrs_
#undef dmvpc_
#undef dnlaso_
#undef dnorcdf_
#undef dnppla_
#undef dnrm2_
#undef dnwla_
#undef dorg2r_
#undef dorgqr_
#undef dorgr2_
#undef dorm2r_
#undef dormqr_
#undef dormr2_
#undef dortqr_
#undef dpmeps_
#undef dpmpar_
#undef dpoco_
#undef dpodi_
#undef dpofa_
#undef dposl_
#undef dqrdc_
#undef dqrsl_
#undef drot_
#undef drotg_
#undef drscl_
#undef dsaitr_
#undef dsapps_
#undef dsaup2_
#undef dsaupd_
#undef dscal_
#undef dsconv_
#undef dseigt_
#undef dsesrt_
#undef dseupd_
#undef dsgets_
#undef dsortr_
#undef dspr_
#undef dsptrf_
#undef dsptrs_
#undef dstats_
#undef dsteqr_
#undef dstqrb_
#undef dsvdc_
#undef dswap_
#undef dtgex2_
#undef dtgexc_
#undef dtgsen_
#undef dtgsy2_
#undef dtgsyl_
#undef dtrmm_
#undef dtrmv_
#undef dtrsl_
#undef dtrsv_
#undef dvsort_
#undef dzasum_
#undef dznrm2_
#undef elmhes_
#undef eltran_
#undef enorm_
#undef epslon_
#undef errclb_
#undef fd_
#undef fdjac2_
#undef formk_
#undef formt_
#undef freev_
#undef fv_
#undef fvd_
#undef fxshfr_
#undef gpfa2f_
#undef gpfa3f_
#undef gpfa5f_
#undef gpfa_
#undef hpsolb_
#undef hqr2_
#undef hqr_
#undef idamax_
#undef ieeeck_
#undef ilaenv_
#undef ins_
#undef isamax_
#undef izamax_
#undef lb1_
#undef lbfgs_
#undef lce_
#undef lde_
#undef lmder1_
#undef lmder_
#undef lmdif_
#undef lmpar_
#undef lnsrlb_
#undef lsame_
#undef lse_
#undef lsqr_
#undef mainlb_
#undef matupd_
#undef mcsrch_
#undef mcstep_
#undef newest_
#undef nextk_
#undef prn1lb_
#undef prn2lb_
#undef prn3lb_
#undef projgr_
#undef pythag_
#undef qrfac_
#undef qrsolv_
#undef quad_
#undef quadit_
#undef quadsd_
#undef realit_
#undef rebak_
#undef reduc_
#undef refine_
#undef rg_
#undef rpoly_
#undef rs_
#undef rsg_
#undef saxpy_
#undef sbeg_
#undef sblas2_
#undef schck1_
#undef schck2_
#undef schck3_
#undef schck4_
#undef schck5_
#undef scnrm2_
#undef scopy_
#undef sdiff_
#undef sdot_
#undef second_
#undef setdgpfa_
#undef setgpfa_
#undef setulb_
#undef sgbmv_
#undef sgemv_
#undef sgeqpf_
#undef sgeqr2_
#undef sger_
#undef sgerq2_
#undef sggsvd_
#undef sggsvp_
#undef simpru_
#undef slacpy_
#undef slags2_
#undef slamc1_
#undef slamc2_
#undef slamc3_
#undef slamc4_
#undef slamc5_
#undef slamch_
#undef slange_
#undef slapll_
#undef slapmt_
#undef slapy2_
#undef slarf_
#undef slarfg_
#undef slartg_
#undef slas2_
#undef slaset_
#undef slassq_
#undef slasv2_
#undef smake_
#undef smessg_
#undef snrm2_
#undef sopen_
#undef sorg2r_
#undef sorm2r_
#undef sormr2_
#undef sqrdc_
#undef sqrsl_
#undef srot_
#undef srotg_
#undef srule_
#undef ssbmv_
#undef sscal_
#undef sspmv_
#undef sspr2_
#undef sspr_
#undef ssvdc_
#undef sswap_
#undef ssymv_
#undef ssyr2_
#undef ssyr_
#undef stbmv_
#undef stbsv_
#undef stgsja_
#undef stpmv_
#undef stpsv_
#undef strmv_
#undef strsv_
#undef subsm_
#undef timer_
#undef tql1_
#undef tql2_
#undef tqlrat_
#undef trapru_
#undef tred1_
#undef tred2_
#undef urand_
#undef xerbla_
#undef xsimpru_
#undef xtrapru_
#undef zaxpy_
#undef zcopy_
#undef zdotc_
#undef zdotu_
#undef zdrot_
#undef zdscal_
#undef zgebak_
#undef zgebal_
#undef zgeev_
#undef zgehd2_
#undef zgehrd_
#undef zgemm_
#undef zgemv_
#undef zgerc_
#undef zhseqr_
#undef zlacgv_
#undef zlacpy_
#undef zladiv_
#undef zlahqr_
#undef zlahrd_
#undef zlange_
#undef zlanhs_
#undef zlarf_
#undef zlarfb_
#undef zlarfg_
#undef zlarft_
#undef zlarfx_
#undef zlascl_
#undef zlaset_
#undef zlassq_
#undef zlatrs_
#undef zqrdc_
#undef zqrsl_
#undef zscal_
#undef zsvdc_
#undef zswap_
#undef ztrevc_
#undef ztrmm_
#undef ztrmv_
#undef ztrsv_
#undef zung2r_
#undef zunghr_
#undef zungqr_
#undef ztgsyl_
#undef zrot_
#undef zlatdf_
#undef zlacn2_
#undef ztgsy2_
#undef ztgexc_
#undef zggbak_
#undef zgetc2_
#undef zhgeqz_
#undef zgges_
#undef zlaswp_
#undef zdrscl_
#undef zlartg_
#undef zggbal_
#undef zgeqr2_
#undef zgecon_
#undef zunmqr_
#undef zunm2r_
#undef zgeqrf_
#undef ztgex2_
#undef zgesc2_
#undef ztgsen_
#undef zgghrd_
#undef zgeru_
#undef izmax1_
#undef dzsum1_
#undef zgees_
#undef ztrexc_
#undef ztrsen_
#undef ztrsyl_
