extern int v3p_netlib_dhgeqz_(
  char *job,
  char *compq,
  char *compz,
  v3p_netlib_integer *n,
  v3p_netlib_integer *ilo,
  v3p_netlib_integer *ihi,
  v3p_netlib_doublereal *a,
  v3p_netlib_integer *lda,
  v3p_netlib_doublereal *b,
  v3p_netlib_integer *ldb,
  v3p_netlib_doublereal *alphar,
  v3p_netlib_doublereal *alphai,
  v3p_netlib_doublereal *beta,
  v3p_netlib_doublereal *q,
  v3p_netlib_integer *ldq,
  v3p_netlib_doublereal *z__,
  v3p_netlib_integer *ldz,
  v3p_netlib_doublereal *work,
  v3p_netlib_integer *lwork,
  v3p_netlib_integer *info,
  v3p_netlib_ftnlen job_len,
  v3p_netlib_ftnlen compq_len,
  v3p_netlib_ftnlen compz_len
  );
