/*=========================================================================
 *
 *  Copyright NumFOCUS
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *         http://www.apache.org/licenses/LICENSE-2.0.txt
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 *=========================================================================*/
#ifndef itkGPUFiniteDifferenceFilterEnum_h
#define itkGPUFiniteDifferenceFilterEnum_h

#include <iostream>
#include "ITKGPUFiniteDifferenceExport.h"

namespace itk
{
/**
 * \class GPUFiniteDifferenceFilterEnum
 * \ingroup ITKGPUFiniteDifference
 */
enum class GPUFiniteDifferenceFilterEnum : uint8_t
{
  UNINITIALIZED,
  INITIALIZED
};
/** Define how to print enumeration values */
extern ITKGPUFiniteDifference_EXPORT std::ostream &
                                     operator<<(std::ostream & out, const GPUFiniteDifferenceFilterEnum value);
} // end namespace itk

#endif
