extern int v3p_netlib_zhseqr_(
  char *job,
  char *compz,
  v3p_netlib_integer *n,
  v3p_netlib_integer *ilo,
  v3p_netlib_integer *ihi,
  v3p_netlib_doublecomplex *h__,
  v3p_netlib_integer *ldh,
  v3p_netlib_doublecomplex *w,
  v3p_netlib_doublecomplex *z__,
  v3p_netlib_integer *ldz,
  v3p_netlib_doublecomplex *work,
  v3p_netlib_integer *lwork,
  v3p_netlib_integer *info,
  v3p_netlib_ftnlen job_len,
  v3p_netlib_ftnlen compz_len
  );
