/*=========================================================================
*
*  Copyright NumFOCUS
*
*  Licensed under the Apache License, Version 2.0 (the "License");
*  you may not use this file except in compliance with the License.
*  You may obtain a copy of the License at
*
*         http://www.apache.org/licenses/LICENSE-2.0.txt
*
*  Unless required by applicable law or agreed to in writing, software
*  distributed under the License is distributed on an "AS IS" BASIS,
*  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*  See the License for the specific language governing permissions and
*  limitations under the License.
*
*=========================================================================*/
// This code fragment is included in the test driver generated by the
// CMake command create_test_sourcelist. This is included directly before
// the call to the test main function.

  vnl_sample_reseed(8775070);
  try
    {
    std::ofstream redirectStream;
    std::streambuf *redirectBuf, *oldCoutBuf = nullptr;
    RedirectOutputParameters& redirectOutputParameters = GetRedirectOutputParameters();
    if (redirectOutputParameters.redirect)
      {
      std::cout << "Test output has been redirected to: " << redirectOutputParameters.fileName << std::endl;
      redirectStream.open(redirectOutputParameters.fileName.c_str(),std::ios::out);
      redirectBuf = redirectStream.rdbuf();
      oldCoutBuf = std::cout.rdbuf();
      std::cout.rdbuf(redirectBuf);
      }
