/*: Factors a double precision symmetric positive definite matrix.  */
extern int v3p_netlib_dpofa_(
  v3p_netlib_doublereal *a,
  v3p_netlib_integer *lda,
  v3p_netlib_integer *n,
  v3p_netlib_integer *info
  );
