extern v3p_netlib_C_f v3p_netlib_cdotc_(
  v3p_netlib_complex * ret_val,
  v3p_netlib_integer *n,
  v3p_netlib_complex *cx,
  v3p_netlib_integer *incx,
  v3p_netlib_complex *cy,
  v3p_netlib_integer *incy
  );
