extern int v3p_netlib_csrot_(
  v3p_netlib_integer *n,
  v3p_netlib_complex *cx,
  v3p_netlib_integer *incx,
  v3p_netlib_complex *cy,
  v3p_netlib_integer *incy,
  v3p_netlib_real *c__,
  v3p_netlib_real *s
  );
