/* mangling test */
#ifndef openjpeg_mangle_h
#define openjpeg_mangle_h

#define allocate_buffers itk_allocate_buffers
#define bio_create itk_bio_create
#define bio_destroy itk_bio_destroy
#define bio_flush itk_bio_flush
#define bio_inalign itk_bio_inalign
#define bio_init_dec itk_bio_init_dec
#define bio_init_enc itk_bio_init_enc
#define bio_numbytes itk_bio_numbytes
#define bio_read itk_bio_read
#define bio_write itk_bio_write
#define dwt_calc_explicit_stepsizes itk_dwt_calc_explicit_stepsizes
#define dwt_decode itk_dwt_decode
#define dwt_decode_1 itk_dwt_decode_1
#define dwt_decode_1_ itk_dwt_decode_1_
#define dwt_decode_real itk_dwt_decode_real
#define dwt_encode itk_dwt_encode
#define dwt_encode_1 itk_dwt_encode_1
#define dwt_encode_1_real itk_dwt_encode_1_real
#define dwt_encode_real itk_dwt_encode_real
#define dwt_getgain itk_dwt_getgain
#define dwt_getgain_real itk_dwt_getgain_real
#define dwt_getnorm itk_dwt_getnorm
#define dwt_getnorm_real itk_dwt_getnorm_real
#define get_all_encoding_parameters itk_get_all_encoding_parameters
#define get_default_stride itk_get_default_stride
#define get_encoding_parameters itk_get_encoding_parameters
#define get_mct_norms itk_get_mct_norms
#define get_mct_norms_real itk_get_mct_norms_real
#define get_tp_stride itk_get_tp_stride
#define j2k_build_decoder itk_j2k_build_decoder
#define j2k_build_encoder itk_j2k_build_encoder
#define j2k_convert_progression_order itk_j2k_convert_progression_order
#define j2k_copy_default_tcp_and_create_tcd itk_j2k_copy_default_tcp_and_create_tcd
#define j2k_create_compress itk_j2k_create_compress
#define j2k_create_decompress itk_j2k_create_decompress
#define j2k_create_tcd itk_j2k_create_tcd
#define j2k_decode itk_j2k_decode
#define j2k_decode_tile itk_j2k_decode_tile
#define j2k_decode_tiles itk_j2k_decode_tiles
#define j2k_decoding_validation itk_j2k_decoding_validation
#define j2k_destroy itk_j2k_destroy
#define j2k_destroy_header_memory itk_j2k_destroy_header_memory
#define j2k_dump_image itk_j2k_dump_image
#define j2k_encode itk_j2k_encode
#define j2k_encoding_validation itk_j2k_encoding_validation
#define j2k_end_compress itk_j2k_end_compress
#define j2k_end_decompress itk_j2k_end_decompress
#define j2k_end_encoding itk_j2k_end_encoding
#define j2k_get_end_header itk_j2k_get_end_header
#define j2k_init_info itk_j2k_init_info
#define j2k_mct_validation itk_j2k_mct_validation
#define j2k_post_write_tile itk_j2k_post_write_tile
#define j2k_read_cbd itk_j2k_read_cbd
#define j2k_read_coc itk_j2k_read_coc
#define j2k_read_cod itk_j2k_read_cod
#define j2k_read_com itk_j2k_read_com
#define j2k_read_crg itk_j2k_read_crg
#define j2k_read_float32_to_float itk_j2k_read_float32_to_float
#define j2k_read_float32_to_int32 itk_j2k_read_float32_to_int32
#define j2k_read_float64_to_float itk_j2k_read_float64_to_float
#define j2k_read_float64_to_int32 itk_j2k_read_float64_to_int32
#define j2k_read_header itk_j2k_read_header
#define j2k_read_header_procedure itk_j2k_read_header_procedure
#define j2k_read_int16_to_float itk_j2k_read_int16_to_float
#define j2k_read_int16_to_int32 itk_j2k_read_int16_to_int32
#define j2k_read_int32_to_float itk_j2k_read_int32_to_float
#define j2k_read_int32_to_int32 itk_j2k_read_int32_to_int32
#define j2k_read_mcc itk_j2k_read_mcc
#define j2k_read_mco itk_j2k_read_mco
#define j2k_read_mct itk_j2k_read_mct
#define j2k_read_plm itk_j2k_read_plm
#define j2k_read_plt itk_j2k_read_plt
#define j2k_read_poc itk_j2k_read_poc
#define j2k_read_ppm itk_j2k_read_ppm
#define j2k_read_ppt itk_j2k_read_ppt
#define j2k_read_qcc itk_j2k_read_qcc
#define j2k_read_qcd itk_j2k_read_qcd
#define j2k_read_rgn itk_j2k_read_rgn
#define j2k_read_siz itk_j2k_read_siz
#define j2k_read_sot itk_j2k_read_sot
#define j2k_read_SPCod_SPCoc itk_j2k_read_SPCod_SPCoc
#define j2k_read_SQcd_SQcc itk_j2k_read_SQcd_SQcc
#define j2k_read_tile_header itk_j2k_read_tile_header
#define j2k_read_tlm itk_j2k_read_tlm
#define j2k_set_decode_area itk_j2k_set_decode_area
#define j2k_setup_decoder itk_j2k_setup_decoder
#define j2k_setup_decoding itk_j2k_setup_decoding
#define j2k_setup_encoder itk_j2k_setup_encoder
#define j2k_setup_header_reading itk_j2k_setup_header_reading
#define j2k_setup_header_writting itk_j2k_setup_header_writting
#define j2k_setup_mct_encoding itk_j2k_setup_mct_encoding
#define j2k_start_compress itk_j2k_start_compress
#define j2k_tcp_destroy itk_j2k_tcp_destroy
#define j2k_update_rates itk_j2k_update_rates
#define j2k_write_coc_in_memory itk_j2k_write_coc_in_memory
#define j2k_write_cod itk_j2k_write_cod
#define j2k_write_com itk_j2k_write_com
#define j2k_write_eoc itk_j2k_write_eoc
#define j2k_write_epc itk_j2k_write_epc
#define j2k_write_float_to_float itk_j2k_write_float_to_float
#define j2k_write_float_to_float64 itk_j2k_write_float_to_float64
#define j2k_write_float_to_int16 itk_j2k_write_float_to_int16
#define j2k_write_float_to_int32 itk_j2k_write_float_to_int32
#define j2k_write_image_components itk_j2k_write_image_components
#define j2k_write_mct_data_group itk_j2k_write_mct_data_group
#define j2k_write_poc itk_j2k_write_poc
#define j2k_write_poc_in_memory itk_j2k_write_poc_in_memory
#define j2k_write_qcc_in_memory itk_j2k_write_qcc_in_memory
#define j2k_write_qcd itk_j2k_write_qcd
#define j2k_write_regions itk_j2k_write_regions
#define j2k_write_siz itk_j2k_write_siz
#define j2k_write_soc itk_j2k_write_soc
#define j2k_write_SQcd_SQcc itk_j2k_write_SQcd_SQcc
#define j2k_write_tile itk_j2k_write_tile
#define j2k_write_tlm itk_j2k_write_tlm
#define j2k_write_updated_tlm itk_j2k_write_updated_tlm
#define jp2_create itk_jp2_create
#define jp2_decode itk_jp2_decode
#define jp2_decode_tile itk_jp2_decode_tile
#define jp2_default_validation itk_jp2_default_validation
#define jp2_destroy itk_jp2_destroy
#define jp2_encode itk_jp2_encode
#define jp2_end_compress itk_jp2_end_compress
#define jp2_end_decompress itk_jp2_end_decompress
#define jp2_read_bpcc itk_jp2_read_bpcc
#define jp2_read_colr itk_jp2_read_colr
#define jp2_read_ftyp itk_jp2_read_ftyp
#define jp2_read_header itk_jp2_read_header
#define jp2_read_header_procedure itk_jp2_read_header_procedure
#define jp2_read_ihdr itk_jp2_read_ihdr
#define jp2_read_jp itk_jp2_read_jp
#define jp2_read_jp2h itk_jp2_read_jp2h
#define jp2_read_tile_header itk_jp2_read_tile_header
#define jp2_set_decode_area itk_jp2_set_decode_area
#define jp2_setup_decoder itk_jp2_setup_decoder
#define jp2_setup_encoder itk_jp2_setup_encoder
#define jp2_skip_jp2c itk_jp2_skip_jp2c
#define jp2_start_compress itk_jp2_start_compress
#define jp2_write_bpcc itk_jp2_write_bpcc
#define jp2_write_colr itk_jp2_write_colr
#define jp2_write_ftyp itk_jp2_write_ftyp
#define jp2_write_ihdr itk_jp2_write_ihdr
#define jp2_write_jp itk_jp2_write_jp
#define jp2_write_jp2c itk_jp2_write_jp2c
#define jp2_write_jp2h itk_jp2_write_jp2h
#define jp2_write_tile itk_jp2_write_tile
#define jpt_init_msg_header itk_jpt_init_msg_header
#define jpt_read_msg_header itk_jpt_read_msg_header
#define jpt_read_VBAS_info itk_jpt_read_VBAS_info
#define jpt_reinit_msg_header itk_jpt_reinit_msg_header
#define mct_decode itk_mct_decode
#define mct_decode_custom itk_mct_decode_custom
#define mct_decode_real itk_mct_decode_real
#define mct_encode itk_mct_encode
#define mct_encode_custom itk_mct_encode_custom
#define mct_encode_real itk_mct_encode_real
#define mct_getnorm itk_mct_getnorm
#define mct_getnorm_real itk_mct_getnorm_real
#define mqc_bypass_enc itk_mqc_bypass_enc
#define mqc_bypass_flush_enc itk_mqc_bypass_flush_enc
#define mqc_bypass_init_enc itk_mqc_bypass_init_enc
#define mqc_create itk_mqc_create
#define mqc_decode itk_mqc_decode
#define mqc_destroy itk_mqc_destroy
#define mqc_encode itk_mqc_encode
#define mqc_erterm_enc itk_mqc_erterm_enc
#define mqc_flush itk_mqc_flush
#define mqc_init_dec itk_mqc_init_dec
#define mqc_init_enc itk_mqc_init_enc
#define mqc_numbytes itk_mqc_numbytes
#define mqc_reset_enc itk_mqc_reset_enc
#define mqc_resetstates itk_mqc_resetstates
#define mqc_restart_enc itk_mqc_restart_enc
#define mqc_restart_init_enc itk_mqc_restart_init_enc
#define mqc_segmark_enc itk_mqc_segmark_enc
#define mqc_setstate itk_mqc_setstate
#define opj_calculate_norms itk_opj_calculate_norms
#define opj_clock itk_opj_clock
#define opj_create_compress itk_opj_create_compress
#define opj_create_decompress itk_opj_create_decompress
#define opj_decode itk_opj_decode
#define opj_decode_tile_data itk_opj_decode_tile_data
#define opj_default_callback itk_opj_default_callback
#define opj_destroy_codec itk_opj_destroy_codec
#define opj_destroy_cstr_info itk_opj_destroy_cstr_info
#define opj_encode itk_opj_encode
#define opj_end_compress itk_opj_end_compress
#define opj_end_decompress itk_opj_end_decompress
#define opj_event_msg itk_opj_event_msg
#define opj_image_comp_update itk_opj_image_comp_update
#define opj_image_create itk_opj_image_create
#define opj_image_create0 itk_opj_image_create0
#define opj_image_destroy itk_opj_image_destroy
#define opj_image_tile_create itk_opj_image_tile_create
#define opj_lupDecompose itk_opj_lupDecompose
#define opj_lupInvert itk_opj_lupInvert
#define opj_lupSolve itk_opj_lupSolve
#define opj_matrix_inversion_f itk_opj_matrix_inversion_f
#define opj_procedure_list_add_procedure itk_opj_procedure_list_add_procedure
#define opj_procedure_list_clear itk_opj_procedure_list_clear
#define opj_procedure_list_create itk_opj_procedure_list_create
#define opj_procedure_list_destroy itk_opj_procedure_list_destroy
#define opj_procedure_list_get_first_procedure itk_opj_procedure_list_get_first_procedure
#define opj_procedure_list_get_nb_procedures itk_opj_procedure_list_get_nb_procedures
#define opj_read_bytes_BE itk_opj_read_bytes_BE
#define opj_read_bytes_LE itk_opj_read_bytes_LE
#define opj_read_double_BE itk_opj_read_double_BE
#define opj_read_double_LE itk_opj_read_double_LE
#define opj_read_float_BE itk_opj_read_float_BE
#define opj_read_float_LE itk_opj_read_float_LE
#define opj_read_from_file itk_opj_read_from_file
#define opj_read_header itk_opj_read_header
#define opj_read_tile_header itk_opj_read_tile_header
#define opj_restrict_decoding itk_opj_restrict_decoding
#define opj_seek_from_file itk_opj_seek_from_file
#define opj_set_decode_area itk_opj_set_decode_area
#define opj_set_default_decoder_parameters itk_opj_set_default_decoder_parameters
#define opj_set_default_encoder_parameters itk_opj_set_default_encoder_parameters
#define opj_set_error_handler itk_opj_set_error_handler
#define opj_set_info_handler itk_opj_set_info_handler
#define opj_set_MCT itk_opj_set_MCT
#define opj_setup_decoder itk_opj_setup_decoder
#define opj_setup_encoder itk_opj_setup_encoder
#define opj_set_warning_handler itk_opj_set_warning_handler
#define opj_skip_from_file itk_opj_skip_from_file
#define opj_start_compress itk_opj_start_compress
#define opj_stream_create itk_opj_stream_create
#define opj_stream_create_default_file_stream itk_opj_stream_create_default_file_stream
#define opj_stream_create_file_stream itk_opj_stream_create_file_stream
#define opj_stream_default_create itk_opj_stream_default_create
#define opj_stream_default_read itk_opj_stream_default_read
#define opj_stream_default_seek itk_opj_stream_default_seek
#define opj_stream_default_skip itk_opj_stream_default_skip
#define opj_stream_default_write itk_opj_stream_default_write
#define opj_stream_destroy itk_opj_stream_destroy
#define opj_stream_flush itk_opj_stream_flush
#define opj_stream_has_seek itk_opj_stream_has_seek
#define opj_stream_read_data itk_opj_stream_read_data
#define opj_stream_read_seek itk_opj_stream_read_seek
#define opj_stream_read_skip itk_opj_stream_read_skip
#define opj_stream_seek itk_opj_stream_seek
#define opj_stream_set_read_function itk_opj_stream_set_read_function
#define opj_stream_set_seek_function itk_opj_stream_set_seek_function
#define opj_stream_set_skip_function itk_opj_stream_set_skip_function
#define opj_stream_set_user_data itk_opj_stream_set_user_data
#define opj_stream_set_write_function itk_opj_stream_set_write_function
#define opj_stream_skip itk_opj_stream_skip
#define opj_stream_tell itk_opj_stream_tell
#define opj_stream_write_data itk_opj_stream_write_data
#define opj_stream_write_seek itk_opj_stream_write_seek
#define opj_stream_write_skip itk_opj_stream_write_skip
#define opj_version itk_opj_version
#define opj_write_bytes_BE itk_opj_write_bytes_BE
#define opj_write_bytes_LE itk_opj_write_bytes_LE
#define opj_write_double_BE itk_opj_write_double_BE
#define opj_write_double_LE itk_opj_write_double_LE
#define opj_write_float_BE itk_opj_write_float_BE
#define opj_write_float_LE itk_opj_write_float_LE
#define opj_write_from_file itk_opj_write_from_file
#define opj_write_tile itk_opj_write_tile
#define pi_check_next_level itk_pi_check_next_level
#define pi_create itk_pi_create
#define pi_create_decode itk_pi_create_decode
#define pi_create_encode itk_pi_create_encode
#define pi_destroy itk_pi_destroy
#define pi_initialise_encode itk_pi_initialise_encode
#define pi_next itk_pi_next
#define pi_update_decode_not_poc itk_pi_update_decode_not_poc
#define pi_update_decode_poc itk_pi_update_decode_poc
#define pi_update_encode_not_poc itk_pi_update_encode_not_poc
#define pi_update_encode_poc_and_final itk_pi_update_encode_poc_and_final
#define pi_update_encoding_parameters itk_pi_update_encoding_parameters
#define _ProfInit itk__ProfInit
#define _ProfPrint itk__ProfPrint
#define _ProfSave itk__ProfSave
#define _ProfStart itk__ProfStart
#define _ProfStop itk__ProfStop
#define raw_create itk_raw_create
#define raw_decode itk_raw_decode
#define raw_destroy itk_raw_destroy
#define raw_init_dec itk_raw_init_dec
#define raw_numbytes itk_raw_numbytes
#define set_default_event_handler itk_set_default_event_handler
#define t1_create itk_t1_create
#define t1_decode_cblks itk_t1_decode_cblks
#define t1_destroy itk_t1_destroy
#define t1_encode_cblks itk_t1_encode_cblks
#define t1_updateflags itk_t1_updateflags
#define t2_create itk_t2_create
#define t2_decode_packets itk_t2_decode_packets
#define t2_destroy itk_t2_destroy
#define t2_encode_packet itk_t2_encode_packet
#define t2_encode_packets itk_t2_encode_packets
#define t2_read_packet_header itk_t2_read_packet_header
#define tcd_code_block_dec_deallocate itk_tcd_code_block_dec_deallocate
#define tcd_code_block_enc_deallocate itk_tcd_code_block_enc_deallocate
#define tcd_copy_tile_data itk_tcd_copy_tile_data
#define tcd_create itk_tcd_create
#define tcd_dc_level_shift_decode itk_tcd_dc_level_shift_decode
#define tcd_dc_level_shift_encode itk_tcd_dc_level_shift_encode
#define tcd_decode_tile itk_tcd_decode_tile
#define tcd_destroy itk_tcd_destroy
#define tcd_dwt_decode itk_tcd_dwt_decode
#define tcd_dwt_encode itk_tcd_dwt_encode
#define tcd_encode_tile itk_tcd_encode_tile
#define tcd_get_decoded_tile_size itk_tcd_get_decoded_tile_size
#define tcd_get_encoded_tile_size itk_tcd_get_encoded_tile_size
#define tcd_init itk_tcd_init
#define tcd_init_decode_tile itk_tcd_init_decode_tile
#define tcd_init_encode_tile itk_tcd_init_encode_tile
#define tcd_makelayer itk_tcd_makelayer
#define tcd_makelayer_fixed itk_tcd_makelayer_fixed
#define tcd_mct_decode itk_tcd_mct_decode
#define tcd_mct_encode itk_tcd_mct_encode
#define tcd_rateallocate itk_tcd_rateallocate
#define tcd_rate_allocate_encode itk_tcd_rate_allocate_encode
#define tcd_rateallocate_fixed itk_tcd_rateallocate_fixed
#define tcd_t1_decode itk_tcd_t1_decode
#define tcd_t1_encode itk_tcd_t1_encode
#define tcd_t2_decode itk_tcd_t2_decode
#define tcd_t2_encode itk_tcd_t2_encode
#define tcd_update_tile_data itk_tcd_update_tile_data
#define tgt_create itk_tgt_create
#define tgt_decode itk_tgt_decode
#define tgt_destroy itk_tgt_destroy
#define tgt_encode itk_tgt_encode
#define tgt_init itk_tgt_init
#define tgt_reset itk_tgt_reset
#define tgt_setvalue itk_tgt_setvalue
#define v4dwt_decode itk_v4dwt_decode
#define dwt_norms itk_dwt_norms
#define dwt_norms_real itk_dwt_norms_real
#define mct_norms itk_mct_norms
#define mct_norms_real itk_mct_norms_real
#define lut_ctxno_zc itk_lut_ctxno_zc
#define lut_ctxno_sc itk_lut_ctxno_sc
#define lut_spb itk_lut_spb
#define lut_nmsedec_sig itk_lut_nmsedec_sig
#define lut_nmsedec_sig0 itk_lut_nmsedec_sig0
#define lut_nmsedec_ref itk_lut_nmsedec_ref
#define lut_nmsedec_ref0 itk_lut_nmsedec_ref0
#define MCT_ELEMENT_SIZE itk_MCT_ELEMENT_SIZE
#define j2k_prog_order_list itk_j2k_prog_order_list
#define mqc_states itk_mqc_states
#define j2k_memory_marker_handler_tab itk_j2k_memory_marker_handler_tab
#define j2k_mct_read_functions_to_int32 itk_j2k_mct_read_functions_to_int32
#define j2k_mct_write_functions_from_float itk_j2k_mct_write_functions_from_float
#define jp2_header itk_jp2_header
#define j2k_mct_read_functions_to_float itk_j2k_mct_read_functions_to_float
#define jp2_img_header itk_jp2_img_header
#define group_list itk_group_list

#endif
