//-*- c++ -*-------------------------------------------------------------------
#ifndef vnl_algo_fwd_h_
#define vnl_algo_fwd_h_


template <class T> struct vnl_matrix_inverse;
template <class T> class vnl_svd;
template <class T> class vnl_qr;
template <class T> class vnl_symmetric_eigensystem;
struct vnl_amoeba_SimplexCorner;
class vnl_cholesky;
class vnl_complex_eigensystem;
class vnl_conjugate_gradient;
class vnl_cpoly_roots;
class vnl_lbfgs;
class vnl_levenberg_marquardt;

#endif // vnl_algo_fwd_h_
