extern int v3p_netlib_eltran_(
  v3p_netlib_integer *nm,
  v3p_netlib_integer *n,
  v3p_netlib_integer *low,
  v3p_netlib_integer *igh,
  v3p_netlib_doublereal *a,
  v3p_netlib_integer *int__,
  v3p_netlib_doublereal *z__
  );
