/*=========================================================================
 *
 *  Copyright NumFOCUS
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *         http://www.apache.org/licenses/LICENSE-2.0.txt
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 *=========================================================================*/
#ifndef itkIOConfigure_h
#define itkIOConfigure_h

#define ITK_SUPPORTS_WCHAR_T_FILENAME_CSTYLEIO
#define ITK_SUPPORTS_WCHAR_T_FILENAME_IOSTREAMS_CONSTRUCTORS
#define ITK_SUPPORTS_FDSTREAM_HPP

/*
 *  Enable the pre-registration of factories for specific image file formats
 */
/* #undef ITKIO_SUPPORTS_ANALYZE_IMAGEIO */
/* #undef ITKIO_SUPPORTS_BIORAD_IMAGEIO */
/* #undef ITKIO_SUPPORTS_BMP_IMAGEIO */
/* #undef ITKIO_SUPPORTS_GDCM_IMAGEIO */
/* #undef ITKIO_SUPPORTS_GIPL_IMAGEIO */
/* #undef ITKIO_SUPPORTS_JPEG_IMAGEIO */
/* #undef ITKIO_SUPPORTS_LSM_IMAGEIO */
/* #undef ITKIO_SUPPORTS_META_IMAGEIO */
/* #undef ITKIO_SUPPORTS_NIFTI_IMAGEIO */
/* #undef ITKIO_SUPPORTS_NRRD_IMAGEIO */
/* #undef ITKIO_SUPPORTS_PNG_IMAGEIO */
/* #undef ITKIO_SUPPORTS_STIMULATE_IMAGEIO */
/* #undef ITKIO_SUPPORTS_TIFF_IMAGEIO */
/* #undef ITKIO_SUPPORTS_VTK_IMAGEIO */

#endif //itkConfigure_h
