extern int v3p_netlib_zlarfb_(
  char *side,
  char *trans,
  char *direct,
  char *storev,
  v3p_netlib_integer *m,
  v3p_netlib_integer *n,
  v3p_netlib_integer *k,
  v3p_netlib_doublecomplex *v,
  v3p_netlib_integer *ldv,
  v3p_netlib_doublecomplex *t,
  v3p_netlib_integer *ldt,
  v3p_netlib_doublecomplex *c__,
  v3p_netlib_integer *ldc,
  v3p_netlib_doublecomplex *work,
  v3p_netlib_integer *ldwork,
  v3p_netlib_ftnlen side_len,
  v3p_netlib_ftnlen trans_len,
  v3p_netlib_ftnlen direct_len,
  v3p_netlib_ftnlen storev_len
  );
